@@ -1,7 +1,7 @@
 ;;;
 ;;; records.el
 ;;;
-;;; $Id: records.el,v 1.27 1999/08/01 02:36:14 ashvin Exp $
+;;; $Id: records.el,v 1.33 2000/01/31 21:23:09 ashvin Exp $
 ;;;
 ;;; Copyright (C) 1996 by Ashvin Goel
 ;;;
@@ -11,12 +11,13 @@
 (require 'records-index)
 (require 'records-dindex)
 (require 'records-util)
+(require 'records-search)
 
 ;;;
 ;;; Internal variables - users shouldn't change
 ;;; The defvar is for internal documentation.
 ;;;
-(defconst records-version "1.4.3")
+(defconst records-version "1.4.8")
 
 (defvar records-mode-menu-map nil
   "Records Menu Map. Internal variable.")
@@ -193,6 +194,21 @@ returns new normalized date."
     ;; return new date
     new-ndate))
 
+(defun records-sane-date (date &optional no-err)
+  "Tests if date is sane. Returns date if so or else error. 
+If no-err is set, returns nil if date is bogus."
+  (if (not (= (length date) records-date-length))
+      (if no-err nil
+        (error (format "records-sane-date: %s: bad date" date)))
+    ;; else
+    (let ((i 0) (ret date))
+      (while (< i records-date-length)
+        (if (or (< (string-to-char (substring date i)) ?0)
+                (> (string-to-char (substring date i)) ?9))
+            (if no-err (setq ret nil)
+              (error (format "records-sane-date: %s: bad date" date))))
+        (setq i (+ i 1))) ret)))
+
 (defun records-file-to-date (&optional file-name)
   "Get the date associated with file-name.
 If file-name is not specified, the current buffers file-name is used."
@@ -203,9 +219,8 @@ If file-name is not specified, the current buffers file-name is used."
 	  (error "records-file-to-date: buffer has no associated file."))
       (setq file-name (file-name-nondirectory buffer-file-name)))
     ;; check that length of file name is meaningful
-    (if (= (length file-name) records-date-length)
-	file-name
-      (error (concat "records-file-to-date: bad file-name: " file-name))))
+    (records-sane-date file-name)
+    file-name)
 
 (defun records-denormalize-date (ndate)
   "Get the file name associated with  date.
@@ -214,7 +229,7 @@ The ndate is normalized and in (day month year) format."
 	(date (make-string records-date-length ? )))
     (if (= records-year-length 2)
 	;; convert to 2 digit year
-	(if (> (nth 2 cdate) 90)
+	(if (< (nth 2 cdate) 2000)
 	    (setcar (nthcdr 2 cdate) (- (nth 2 cdate) 1900))
 	  (setcar (nthcdr 2 cdate) (- (nth 2 cdate) 2000))))
     ;; now denormalize
@@ -226,7 +241,8 @@ The ndate is normalized and in (day month year) format."
 		(len (nth 2 x)))
 	    (setq date (concat 
 			(substring date 0 start)
-			(format (concat "%0" (int-to-string len) "d") (nth i ndate))
+			(format (concat "%0" (int-to-string len) "d") 
+                                (nth i ndate))
 			(substring date (+ start len))))
 	    (setq i (1+ i))))
        records-date))
@@ -234,6 +250,7 @@ The ndate is normalized and in (day month year) format."
 
 (defun records-normalize-date (date)
   "Returns date in (day month year) format with year in four digits"
+  (records-sane-date date)
   (let ((ndate '(0 0 0))
 	(i 0))
     (mapcar
@@ -272,29 +289,44 @@ With absolute set, get the absolute path."
          (error "records-directory-path: bad records-directory-structure value"))))
 
 (defun records-read-subject (&optional subject)
-  "Read the records subject to be inserted from the minibuffer.
+  "Read the records subject from minibuffer.
 Completion is possible."
   (interactive
    (progn (records-index-buffer); initializes records-subject-table if required
 	  (list (completing-read "Records subject: " records-subject-table))))
   subject)
 
+(defun records-read-date ()
+  "Reads the records date from minibuffer.
+Ensure date is reasonable."
+  (interactive)
+  (let ((date (read-from-minibuffer "Records date: "
+                                    (cons (records-file-to-date) 0))))
+    (while (not (records-sane-date date t)) ; check for date sanity
+      (message (format "%s: bad date" date))
+      (sleep-for 1)
+      (setq date (read-from-minibuffer "Records date: "
+                                       (cons (records-file-to-date) 0))))
+    date))
+
 (defun records-add-text-properties (beg end)
-  "Fontify a records region, make read-only etc.
-Look at variables records-fontify and records-subject-read-only.  
-This function is currently only invoked for a records subject.  
+  "Make a region read-only, etc.
+Look at variable records-subject-read-only.
+This function is currently only invoked for a records subject.
 
 Although the region is read-only, it is possible to edit at the beginning of
 the subject. This can mess up a records subject if anything but a newline is
 inserted. We could close the beginning of the region (see start-close), but
 then users would not be able to add newlines before a subject, and it screws up
-records-encrypt-record and records-decrypt-record. What we need is that 
+records-encrypt-record and records-decrypt-record. What we need is that
 insertion of any character automatically inserts a newline also. TODO"
-  (if records-fontify
-      (progn
-	(add-text-properties beg end '(face bold start-open t))
-	(if records-subject-read-only
-	    (add-text-properties beg end '(read-only records-subject))))))
+  (progn
+    (if (not running-xemacs)
+        ;; emacs has an of-by-one error
+        (setq end (1- end)))
+    (add-text-properties beg end '(start-open t))
+    (if records-subject-read-only
+        (add-text-properties beg end '(read-only records-subject)))))
 
 (defun records-remove-text-properties (s) 
   "Remove the text properties of string in a record.
@@ -431,11 +463,7 @@ ring."
     (insert "* " subject "\n")
     (insert-char ?- (+ (length subject) 2))
     (insert (concat "\n" (records-make-link subject date tag) "\n"))
-    (if running-xemacs
-        (records-add-text-properties opoint (point))
-      ;; emacs has an of-by-one error
-        (records-add-text-properties opoint (1- (point)))
-        )
+    (records-add-text-properties opoint (point))
     (if record-body
 	(insert record-body))))
 
@@ -470,8 +498,10 @@ With arg., keep the body and remove the subject only."
       (insert "\n" (buffer-substring bol bospaces))
       (insert-char ?- (- eol bospaces)))))
 
-;; Thanks to Kaarthik Sivakumar, 04/13/99 for the http, ftp, mailto 
-;; and gopher handling code
+;; 04/13/1999: http, ftp, mailto and gopher handling code
+;;             Thanks to Kaarthik Sivakumar
+;; 01/10/2000: dejanews handling code
+;;             Thanks to Robert Mihram 
 (defun records-goto-link ()
   "Goto the link around point in the records file.
 A link can be any of the following. They must be enclosed in <>.
@@ -483,6 +513,7 @@ A tag is a number.
 5. http:// or mailto:// or ftp:// gopher://  
    The last case is handled by browse-url-browser-function. 
    Refer to Options/\"Open URL with\" in XEmacs. 
+6. <message-id> fetch a article using deja.com.
    Spaces and other funky characters in the url can break this code."
   (interactive)
   (save-excursion
@@ -516,14 +547,25 @@ A tag is a number.
      ((looking-at "<\\(\\(http\\|mailto\\|ftp\\|gopher\\):[^>]+\\)>")
       (funcall browse-url-browser-function 
                (buffer-substring-no-properties (match-beginning 1) 
-                                               (match-end 1))))
+					       (match-end 1))))
+     ((looking-at "<\\([^ \t\n>]+\\)>")
+      (funcall browse-url-browser-function 
+	       (concat
+		"http://search.dejanews.com/msgid.xp?MID=%3C"
+		(buffer-substring-no-properties (match-beginning 1) 
+						(match-end 1))
+		"%3E&format=threaded")))
      (t (error "records-goto-link: invalid link under point.")))))
 
 (defun records-goto-mouse-link (e)
-  "Goto the link where mouse is clicked."
+  "When mouse is clicked on a link, goto the link. 
+When mouse is clicked anywhere else, invoke the default mouse binding."
   (interactive "e")
-  (mouse-set-point e)
-  (records-goto-link))
+  (condition-case nil
+      (progn (mouse-set-point e)
+             (records-goto-link))
+    (error (funcall (global-key-binding [(button2)]) e)))
+  )
 
 (defun records-goto-record (subject date tag 
 				&optional no-hist no-switch todo no-error dir)
@@ -660,9 +702,10 @@ See also records-goto-next-record-file."
 (defun records-goto-today ()
   "Go to the records file of today."
   (interactive)
-    (records-goto-record nil (records-todays-date) "" nil nil records-todo-today))
+    (records-goto-record nil (records-todays-date) "" nil nil
+                         records-todo-today))
 
-(defun records-goto-relative-record-file(&optional arg no-switch no-error)
+(defun records-goto-relative-record-file (&optional arg no-switch no-error)
   "With positive arg, go arg files ahead of current records file. 
 With negative arg, go arg files behind of current records file.
 Returns the new date."
@@ -752,11 +795,12 @@ Identical record files are not put in the history consecutively."
   (let* ((subject (if subject
 		      subject (call-interactively 'records-read-subject)))
 	 (date (records-file-to-date))
-	 (tag ""))
-    ;; we don't currently allow a record insertion 
-    ;; if another record with the same subject exists below this record.
+	 (tag "")
+         (prev-modified (buffer-modified-p)))
+    ;; currently, we don't allow a record insertion if another record with 
+    ;; the same subject exists below this record.
     (save-excursion
-      (if (records-goto-down-record subject)
+      (if (records-goto-down-record subject t)
 	  (error 
 	   (concat "records-insert-record: can't insert out-of-order record: "
 		   subject))))
@@ -775,7 +819,19 @@ Identical record files are not put in the history consecutively."
     (records-dindex-insert-record date)
 
     ;; now make the record body
-    (records-make-record subject date tag record-body)))
+    (records-make-record subject date tag record-body)
+    (if (not prev-modified)
+        (save-buffer))
+    ))
+
+(defun records-delete-empty-record-file ()
+  "Delete current buffer and its associated file if the buffer is empty 
+and not modified."
+  (if (and (not (buffer-modified-p)) (eq (buffer-size) 0))
+      (let* ((bname (buffer-name))
+             (dir-name (records-directory-path bname t)))
+        (kill-buffer nil)
+        (delete-file (concat dir-name "/" bname)))))
 
 (defun records-delete-record (&optional keep-body no-prompt)
   "Delete the current record for the current date.
@@ -784,17 +840,23 @@ With arg, removes the subject only."
   (let* ((date (records-file-to-date))
 	 (subject-tag (records-subject-tag t))
 	 (subject (nth 0 subject-tag))
-	 (tag (nth 1 subject-tag)))
+	 (tag (nth 1 subject-tag))
+         (prev-modified (buffer-modified-p)))
     
     (if (if no-prompt     ;; prompt?
 	    t (y-or-n-p (concat "Delete record: " subject " ")))
 	(progn
 	  ;; remove the record subject and optionally the body
 	  (records-free-record keep-body)
+          (if (not prev-modified)
+              (save-buffer))
 	  ;; remove the date from the date-index
 	  (records-dindex-delete-record date)
 	  ;; remove the records index entry
-	  (records-index-delete-record subject date tag)))
+	  (records-index-delete-record subject date tag)
+          ;; remove empty record and file
+          (records-delete-empty-record-file)
+          ))
     ))
 	
 (defun records-rename-record ()
@@ -803,6 +865,23 @@ With arg, removes the subject only."
   (records-delete-record 'keep-body)
   (records-insert-record))
 
+(defun records-move-record (&optional date)
+  "Move the current record to date. Prompts for date."
+  (interactive)
+  (if (null date)
+      (setq date (records-read-date))
+    (records-sane-date date))
+  (let* ((subject-tag (records-subject-tag t))
+         (subject (nth 0 subject-tag))
+         (point-pair (records-record-region t))
+         (record-body 
+          (buffer-substring (first point-pair) (second point-pair))))
+    (save-excursion
+      (records-goto-record nil date "" nil t nil)
+      (goto-char (point-max))
+      (records-insert-record subject record-body)))
+  (records-delete-record nil t))
+
 (define-derived-mode records-mode text-mode "Records"
   "Enable records-mode for a buffer. Currently, the documentation of this 
 mode exists in three places: the INSTALL and README files and the menubar!
@@ -817,7 +896,8 @@ The key-bindings of this mode are:
 
   (define-key records-mode-map "\C-c\C-i" 'records-insert-record)
   (define-key records-mode-map "\C-c\C-d" 'records-delete-record)
-  (define-key records-mode-map "\C-c\C-r" 'records-rename-record)
+  (define-key records-mode-map "\C-c\C-e" 'records-rename-record)
+  (define-key records-mode-map "\C-c\C-m" 'records-move-record)
 
   (define-key records-mode-map "\M-\C-a" 'records-goto-up-record)
   (define-key records-mode-map "\M-\C-e" 'records-goto-down-record)
@@ -825,29 +905,31 @@ The key-bindings of this mode are:
   (define-key records-mode-map "\C-c\C-p" 'records-goto-prev-record)
   (define-key records-mode-map "\C-c\C-n" 'records-goto-next-record)
 
-  (define-key records-mode-map "\C-c\C-y" 'records-goto-prev-day); yesterday
-  (define-key records-mode-map "\C-c\C-t" 'records-goto-next-day); tomorrow
   (define-key records-mode-map "\C-c\C-b" 'records-goto-prev-record-file)
                                         ; back file
   (define-key records-mode-map "\C-c\C-f" 'records-goto-next-record-file) 
                                         ; front file
 
+  (define-key records-mode-map "\C-c\C-y" 'records-goto-prev-day); yesterday
+  (define-key records-mode-map "\C-c\C-t" 'records-goto-next-day); tomorrow
+
   (define-key records-mode-map "\C-c\C-g" 'records-goto-link)
   (define-key records-mode-map "\C-c\C-l" 'records-goto-last-record)
   (define-key records-mode-map "\C-c\C-j" 'records-goto-index); jump!!
 
-  ;; (define-key records-mode-map [M-S-mouse-1] 'records-goto-mouse-link)
+  (define-key records-mode-map "\C-c\C-s" 'records-search-forward)
+  (define-key records-mode-map "\C-c\C-r" 'records-search-backward)
+
+  (define-key records-mode-map [(button2)] 'records-goto-mouse-link)
 
   ;; utility functions have C-c/ prefix keys
-  (define-key records-mode-map "\C-c/e" 'records-encrypt-record)
-  (define-key records-mode-map "\C-c/d" 'records-decrypt-record)
-
   (define-key records-mode-map "\C-c/t" 'records-create-todo)
   (define-key records-mode-map "\C-c/g" 'records-get-todo)
+  (define-key records-mode-map "\C-c/e" 'records-encrypt-record)
+  (define-key records-mode-map "\C-c/d" 'records-decrypt-record)
   (define-key records-mode-map "\C-c/c" 'records-concatenate-records)
   (define-key records-mode-map "\C-c/f" 'records-concatenate-record-files)
 
-
   (define-key records-mode-map "\C-c\C-c" 'records-goto-calendar)
   (define-key records-mode-map "\C-c\C-k" 'records-link-as-kill)
   (define-key records-mode-map [?\C-c ?\C--] 'records-underline-line)
@@ -858,6 +940,11 @@ The key-bindings of this mode are:
       ()
     (setq records-mode-menu-map
 	  '(["Today's Record" records-goto-today t]
+	    "--"
+	    ["Insert Record" records-insert-record t]
+	    ["Delete Record" records-delete-record t]
+	    ["Rename Record" records-rename-record t]
+	    ["Move Record" records-move-record t]
 	    "--"
 	    ["Up Record" records-goto-up-record t]
 	    ["Down Record" records-goto-down-record t]
@@ -875,9 +962,8 @@ The key-bindings of this mode are:
 	    ["Goto Last Record" records-goto-last-record t]
 	    ["Goto Index" records-goto-index t]
 	    "--"
-	    ["Insert Record" records-insert-record t]
-	    ["Delete Record" records-delete-record t]
-	    ["Rename Record" records-rename-record t]
+            ["Search Forward" records-search-forward t]
+            ["Search Backward" records-search-backward t]
 	    "--"
 	    ["Create TODO" records-create-todo t]
 	    ["Get TODO's" records-get-todo t]
@@ -907,7 +993,7 @@ The key-bindings of this mode are:
   ;; imenu stuff 
   (if (locate-library "imenu")
       (progn
-	(require 'imenu)
+	(eval-when-compile (require 'imenu))
 	(make-variable-buffer-local 'imenu-prev-index-position-function)
 	(make-variable-buffer-local 'imenu-extract-index-name-function)
 	(setq imenu-prev-index-position-function 'records-goto-up-record)
@@ -920,8 +1006,19 @@ The key-bindings of this mode are:
       ()
     (records-initialize)
     (setq records-initialize t))
+  ;; fontification code by Robert Mihram
+  (if (and (or (not (boundp 'font-lock-auto-fontify)) 
+               (not font-lock-auto-fontify))
+           records-mode-use-font-lock)
+      (progn (eval-when-compile (require 'font-lock))
+             (make-local-variable 'font-lock-defaults)
+             (setq font-lock-defaults '(records-mode-font-lock-keywords))
+             (font-lock-mode 1)))
   (run-hooks 'records-mode-hooks)
   )
 
+;; for xemacs
+(put 'records-mode 'font-lock-defaults '(records-mode-font-lock-keywords))
+
 (run-hooks 'records-load-hooks)
 (provide 'records)
